/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2002, 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   handles command loop
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <errno.h>

#include "ecurses.h"
#include "cmd.h"
#include "status.h"
#include "folder.h"
#include "mailreader.h"
#include "screen.h"
#include "error.h"
#include "elmo.h"
#include "box_selection.h"
#include "xmalloc.h"
#include "fetch.h"
#include "confread.h"
#include "compose.h"
#include "sender.h"
#include "attach.h"
#include "clock.h"
#include "abook.h"
#include "read.h"
#include "exec.h"
#include "keymap.h"
#include "hook.h"
#include "lcs.h"
#include "choose.h"
#include "ask.h"
#include "bayes.h"
#include "gettext.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define STACK_SIZE 10

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static cmd_state_t  state_stack[STACK_SIZE];
static cmd_state_t *state_top;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/

WINDOW *cmd_echo_area = NULL;

keymap_t keymaps[CMD_STATE_COUNT];
hook_t   state_back_hooks[CMD_STATE_COUNT];
hook_t   state_leave_hooks[CMD_STATE_COUNT];

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/

static void destroy_keymaps (void);

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static void
quit (void)
{
  delwin (cmd_echo_area);
  destroy_keymaps ();
  elmo_finish (0);
  exit (EXIT_SUCCESS);
}



static void
init_keymaps (void)
{
  int     i;

  for (i = 0; i < CMD_STATE_COUNT; i++)
    keymap_init (keymaps + i);

  keymap_add (keymaps + CMD_LIST, KEY_PPAGE, folder_page_prev);
  keymap_add (keymaps + CMD_LIST, KEY_NPAGE, folder_page_next);
  keymap_add (keymaps + CMD_LIST, KEY_HOME,  folder_bar_first);
  keymap_add (keymaps + CMD_LIST, KEY_END,   folder_bar_last);
  keymap_add (keymaps + CMD_LIST, KEY_UP,    folder_bar_prev);
  keymap_add (keymaps + CMD_LIST, KEY_DOWN,  folder_bar_next);
  keymap_add (keymaps + CMD_LIST, KEY_DC,    folder_delete_mail);
  keymap_add (keymaps + CMD_LIST, '$',       folder_flush);
  keymap_add (keymaps + CMD_LIST, 'd',       folder_sort_date);
  keymap_add (keymaps + CMD_LIST, 't',       folder_sort_threads);
  keymap_add (keymaps + CMD_LIST, 'A',       folder_sort_from);
  keymap_add (keymaps + CMD_LIST, 's',       folder_sort_subject);
  keymap_add (keymaps + CMD_LIST, 'i',       folder_sort_index);
  keymap_add (keymaps + CMD_LIST, 'a',       abook_open);
  keymap_add (keymaps + CMD_LIST, 'q',       cmd_quit);
  keymap_add (keymaps + CMD_LIST, '\r',      mailreader_init);
  keymap_add (keymaps + CMD_LIST, 'F',       fetch_open);
  keymap_add (keymaps + CMD_LIST, 'm',       sender_open_new);
  keymap_add (keymaps + CMD_LIST, 'r',       sender_open_reply);
  keymap_add (keymaps + CMD_LIST, 'f',       sender_open_fwd);
  keymap_add (keymaps + CMD_LIST, 'c',       box_selection_open);
  keymap_add (keymaps + CMD_LIST, ' ',       folder_toggle_flag);

  keymap_add (keymaps + CMD_SELECT_BOX, KEY_PPAGE, box_selection_prev_page);
  keymap_add (keymaps + CMD_SELECT_BOX, KEY_NPAGE, box_selection_next_page);
  keymap_add (keymaps + CMD_SELECT_BOX, KEY_HOME,  box_selection_first);
  keymap_add (keymaps + CMD_SELECT_BOX, KEY_END,   box_selection_last);
  keymap_add (keymaps + CMD_SELECT_BOX, KEY_UP,    box_selection_prev);
  keymap_add (keymaps + CMD_SELECT_BOX, KEY_DOWN,  box_selection_next);
  keymap_add (keymaps + CMD_SELECT_BOX, 'q',       box_selection_leave);
  keymap_add (keymaps + CMD_SELECT_BOX, '\r',      box_selection_hit);
  
  keymap_add (keymaps + CMD_ABOOK, KEY_PPAGE, abook_prev_page);
  keymap_add (keymaps + CMD_ABOOK, KEY_NPAGE, abook_next_page);
  keymap_add (keymaps + CMD_ABOOK, KEY_HOME,  abook_first);
  keymap_add (keymaps + CMD_ABOOK, KEY_END,   abook_last);
  keymap_add (keymaps + CMD_ABOOK, KEY_DOWN,  abook_next);
  keymap_add (keymaps + CMD_ABOOK, KEY_UP,    abook_prev);
  keymap_add (keymaps + CMD_ABOOK, KEY_DC,    abook_remove);
  keymap_add (keymaps + CMD_ABOOK, 'n',       abook_change_name);
  keymap_add (keymaps + CMD_ABOOK, 'e',       abook_change_email);
  keymap_add (keymaps + CMD_ABOOK, 'f',       abook_change_flag);
  keymap_add (keymaps + CMD_ABOOK, 'a',       abook_insert);
  keymap_add (keymaps + CMD_ABOOK, 'q',       abook_close);

  keymap_add (keymaps + CMD_READ_MAIL, 'q',       mailreader_close);
  keymap_add (keymaps + CMD_READ_MAIL, KEY_PPAGE, mailreader_page_up);
  keymap_add (keymaps + CMD_READ_MAIL, KEY_NPAGE, mailreader_page_down);
  keymap_add (keymaps + CMD_READ_MAIL, KEY_HOME,  mailreader_top);
  keymap_add (keymaps + CMD_READ_MAIL, KEY_END,   mailreader_bottom);
  keymap_add (keymaps + CMD_READ_MAIL, KEY_DOWN,  mailreader_line_down);
  keymap_add (keymaps + CMD_READ_MAIL, KEY_UP,    mailreader_line_up);
  keymap_add (keymaps + CMD_READ_MAIL, 'a',       mailreader_attach);
  keymap_add (keymaps + CMD_READ_MAIL, 'b',       abook_add_open);

  keymap_add (keymaps + CMD_FETCH, KEY_PPAGE, fetch_prev_page);
  keymap_add (keymaps + CMD_FETCH, KEY_NPAGE, fetch_next_page);
  keymap_add (keymaps + CMD_FETCH, KEY_HOME,  fetch_first);
  keymap_add (keymaps + CMD_FETCH, KEY_END,   fetch_last);
  keymap_add (keymaps + CMD_FETCH, KEY_DOWN,  fetch_next);
  keymap_add (keymaps + CMD_FETCH, KEY_UP,    fetch_prev);
  keymap_add (keymaps + CMD_FETCH, 'd',       fetch_del_mail);
  keymap_add (keymaps + CMD_FETCH, 'f',       fetch_get_mail);
  keymap_add (keymaps + CMD_FETCH, 'q',       fetch_close);
  keymap_add (keymaps + CMD_FETCH, 'r',       fetch_rset);

  keymap_add (keymaps + CMD_ATTACH, KEY_PPAGE, attach_prev_page);
  keymap_add (keymaps + CMD_ATTACH, KEY_NPAGE, attach_next_page);
  keymap_add (keymaps + CMD_ATTACH, KEY_HOME,  attach_first);
  keymap_add (keymaps + CMD_ATTACH, KEY_END,   attach_last);
  keymap_add (keymaps + CMD_ATTACH, KEY_DOWN,  attach_next);
  keymap_add (keymaps + CMD_ATTACH, KEY_UP,    attach_prev);
  keymap_add (keymaps + CMD_ATTACH, 's',       attach_save);
  keymap_add (keymaps + CMD_ATTACH, 'q',       attach_close);
  keymap_add (keymaps + CMD_ATTACH, '\r',      attach_hit);

  keymap_add (keymaps + CMD_ABOOK_ADD, KEY_PPAGE, abook_add_prev_page);
  keymap_add (keymaps + CMD_ABOOK_ADD, KEY_NPAGE, abook_add_next_page);
  keymap_add (keymaps + CMD_ABOOK_ADD, KEY_HOME,  abook_add_first);
  keymap_add (keymaps + CMD_ABOOK_ADD, KEY_END,   abook_add_last);
  keymap_add (keymaps + CMD_ABOOK_ADD, KEY_DOWN,  abook_add_next);
  keymap_add (keymaps + CMD_ABOOK_ADD, KEY_UP,    abook_add_prev);
  keymap_add (keymaps + CMD_ABOOK_ADD, 'q',       abook_add_close);
  keymap_add (keymaps + CMD_ABOOK_ADD, '\r',      abook_add_hit);

  keymap_add (keymaps + CMD_SENDER, KEY_PPAGE, sender_prev_page);
  keymap_add (keymaps + CMD_SENDER, KEY_NPAGE, sender_next_page);
  keymap_add (keymaps + CMD_SENDER, KEY_HOME,  sender_first);
  keymap_add (keymaps + CMD_SENDER, KEY_END,   sender_last);
  keymap_add (keymaps + CMD_SENDER, KEY_DOWN,  sender_next);
  keymap_add (keymaps + CMD_SENDER, KEY_UP,    sender_prev);
  keymap_add (keymaps + CMD_SENDER, 'f',       sender_change_from);
  keymap_add (keymaps + CMD_SENDER, 's',       sender_change_subject);
  keymap_add (keymaps + CMD_SENDER, 'r',       sender_change_reply_to);
  keymap_add (keymaps + CMD_SENDER, 'a',       sender_add_attachment);
  keymap_add (keymaps + CMD_SENDER, KEY_DC,    sender_delete_attachment);
  keymap_add (keymaps + CMD_SENDER, 'd',       sender_delete_attachment);
  keymap_add (keymaps + CMD_SENDER, 20,        sender_change_content_type);
  keymap_add (keymaps + CMD_SENDER, 'q',       sender_close);
  keymap_add (keymaps + CMD_SENDER, 'y',       sender_go);
  keymap_add (keymaps + CMD_SENDER, 'E',       sender_reedit);

  keymap_add (keymaps + CMD_ASK, KEY_HOME, choose_first);
  keymap_add (keymaps + CMD_ASK, KEY_END,  choose_last);
  keymap_add (keymaps + CMD_ASK, KEY_DOWN, choose_next);
  keymap_add (keymaps + CMD_ASK, KEY_UP,   choose_prev);

  keymap_add (keymaps + CMD_LIST, '~', cmd_test_action);
}


static void
destroy_keymaps (void)
{
  int i;

  for (i = 0; i < CMD_STATE_COUNT; i++){
    keymap_destroy (keymaps + i);
  }
}


static void
init (void)
{
  cmd_echo_area = newwin (1, COLS - 13, LINES - 1, 0);
  keypad (cmd_echo_area, TRUE);
  meta (cmd_echo_area, TRUE);
  cbreak ();
  init_keymaps ();

  state_top  = state_stack;
  *state_top = CMD_LIST;

  hook_add (state_back_hooks + CMD_LIST, folder_soft_refresh);
  hook_add (state_back_hooks + CMD_READ_MAIL, mailreader_refresh);

  hook_add (state_leave_hooks + CMD_READ_MAIL, folder_after_read);
  hook_add (state_leave_hooks + CMD_FETCH, folder_re_read);
  hook_add (state_leave_hooks + CMD_ATTACH, mailreader_change_mime);
  hook_add (state_leave_hooks + CMD_SENDER, compose_after_send);
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/


void
cmd_init (void)
{
  init ();
}



void
cmd_quit (void)
{
  quit ();
}



void
cmd_state_push (cmd_state_t requested_state)
{
  if (state_top == state_stack + STACK_SIZE){
    error_ (1, 0, _("state stack overflow, please send bug report"));
  }
  state_top++;
  *state_top = requested_state;
}



void
cmd_state_pop (void)
{
  if (state_top == state_stack){
    error_ (1, 0, _("state stack underflow, please send bug report"));
  }
  state_top--;

  hook_execute (state_leave_hooks + state_top[1]);
  hook_execute (state_back_hooks + *state_top);
}



cmd_state_t
cmd_state_get (void)
{
  return *state_top;
}



void
cmd_refresh (void)
{
  switch (*state_top){

    case CMD_LIST:
      folder_soft_refresh ();
      break;

    case CMD_READ_MAIL:
      mailreader_refresh ();
      break;

    case CMD_SENDER:
      sender_refresh ();
      break;

    default:
      break;
  }
}


void
cmd_read_loop (void)
{
  int   c;
  char *arg;
  
  while (1){

    screen_doupdate ();

    c = wgetch (cmd_echo_area);
    if (c == ':'){
      arg = read_argument (": ", NULL);
      if (arg)
        exec_run_name (arg);
      continue;
    }
    keymap_action (keymaps + *state_top, c);
  }
}


void
cmd_choose (void)
{
  int c;
  
  while (1){

    screen_doupdate ();
    c = wgetch (cmd_echo_area);
    if (c == '\r'){
      break;
    }
    if (c == 7){
      choose_cancel ();
      break;
    }
    keymap_action (keymaps + CMD_ASK, c);
  }
  hook_execute (state_back_hooks + *state_top);
}



void
cmd_test_action (void)
{
#ifdef XMALLOC_DEBUG


  
#endif
}

/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE cmd.c
 *
 ****************************************************************************/
