/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2002, 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   This module acts as parser of config file.  It reads given file
   and puts pairs (key, value) in hashing table, so they are easy
   to look up later.  All you need to specify is file name, and
   supposed number of pairs.

*/

#define _GNU_SOURCE 1

/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>

#ifdef HAVE_WORDEXP_H
# include <wordexp.h>
#endif

#include "misc.h"
#include "xmalloc.h"
#include "hash.h"
#include "error.h"
#include "confread.h"
#include "file.h"
#include "elmo.h"
#include "cmd.h"
#include "exec.h"
#include "rstring.h"
#include "debug.h"
#include "gettext.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define BUFFER_SIZE 128
#define CONF_FILE ".elmorc"

enum type_t {TYPE_TXT, TYPE_NUM};

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/

struct conf_t {
  enum type_t type;
  union {
    char *txt;
    int   num;
  } value;
};

/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static htable_t *conf_table;
static int       initialized = 0;
static char     *config_filename;

static htable_t *list_table;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/

static int  fill_buffer (char *buf, unsigned size, FILE *fp);
static void add_new_variable (char *line);
static void process_line (char *buf);
static void destroy_conf_item (void *item);
  
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

/*
  This procedure assumes that everything in buffer till last newline
  has already been processed.  It will be overwritten with new data.
  This procedure assumes also that it will not be called again after
  it returned -1.

  buf is always null terminated, last byte is never filled
  buf must not be changed in mean time

*/

static int
fill_buffer (char *buf, unsigned size, FILE *fp){
  static int  position = -1; /* of \n in buf */
  int         result;
  int         ret;
  char       *tmp;

  if (position == -1){
    result = fread (buf, 1, BUFFER_SIZE - 1, fp);
  }
  else {
    memmove (buf, buf + position + 1, BUFFER_SIZE - position - 2);
    ret = fread (buf + BUFFER_SIZE - position - 2, 1,
                 position + 1, fp);

    result = ret + BUFFER_SIZE - position - 2;
  }

  tmp = strrchr (buf, '\n');
  if (!tmp)
    return -1;

  position = tmp - buf;
  return result;
}



static char *
expand (char *str)
{
#ifdef HAVE_WORDEXP_H
  wordexp_t  word_vector;
  int        len = 0;
  int        i;
  char      *result;
  char      *end;

  if (wordexp (str, &word_vector, 0)){
    return NULL;
  }

  for (i = 0; i < word_vector.we_wordc; i++){
    len += strlen (word_vector.we_wordv[i]);
  }

  result = xmalloc (len + 1);
  end    = result;
  for (i = 0; i < word_vector.we_wordc; i++){
    end  = stpcpy (end, word_vector.we_wordv[i]);
  }
  *end   = '\0';
  wordfree (&word_vector);

  return result;
#else
  return xstrdup (str);
#endif
}



static void
add_new_variable (char *line){
  entry_t       *entry;
  char          *tmp;
  struct conf_t *new;
  int            numval = -1;
  char          *rest   = line;
  
  tmp = strchr (line, ' ');
  if (tmp == NULL)
    tmp = strchr (line, '\t');

  if (tmp){
    *tmp = '\0';    /* it is replaced back with space at end of function */
    tmp++;
    if (isdigit (*tmp))
      numval = strtol (tmp, &rest, 0);
  }
  else
    return;

  new = xmalloc (sizeof (struct conf_t));
  if (numval != -1 && !*rest){
    new->type      = TYPE_NUM;
    new->value.num = numval;
  }
  else {
    new->type      = TYPE_TXT;
    new->value.txt = expand (tmp);
  }
  entry = htable_insert (conf_table, line, new);

  if (entry->content != new)
    destroy_conf_item (new);

  if (tmp)
    *(tmp - 1) = ' '; /* tmp++ */
}



static cmd_state_t
name_to_state (char *name)
{
  if (!strcmp (name, "folder"))
    return CMD_LIST;
  if (!strcmp (name, "mail"))
    return CMD_READ_MAIL;
  if (!strcmp (name, "select"))
    return CMD_SELECT_BOX;
  if (!strcmp (name, "fetch"))
    return CMD_FETCH;
  if (!strcmp (name, "send"))
    return CMD_SENDER;
  if (!strcmp (name, "attach"))
    return CMD_ATTACH;
  if (!strcmp (name, "abook"))
    return CMD_ABOOK;
  if (!strcmp (name, "abook_add"))
    return CMD_ABOOK_ADD;
  if (!strcmp (name, "ask"))
    return CMD_ASK;
  return CMD_INVALID;
}



static int
parse_key (char *txt)
{
  if (isdigit (*txt))
    return atoi (txt);
  return *txt;
}



static void
add_new_key (char *line)
{
  char        *seek = strchr (line, ' ');
  cmd_state_t  state;
  int          key;
  exec_t      *exec;

  if (seek == NULL)
    seek = strchr (line, '\t');
  
  if (seek){
    *seek = '\0';
  }
  else
    return;

  state = name_to_state (line);
  if (state == CMD_INVALID){
    error_ (0, 0, _("invalid state in key definition '%s'"), line);
    return;
  }

  *seek = ' ';
  key   = parse_key (seek + 1);
  seek  = strchr (seek + 1, ' ');
  if (seek == NULL)
    return;

  exec = exec_lookup (seek + 1);
  if (exec == NULL){
    error_ (0, 0, _("invalid function name '%s'"), seek + 1);
    return;
  }

  keymap_add (keymaps + state, key, exec->fun);
}



static void
add_new_hook (char *line)
{
  char   *seek = strchr (line, ' ');
  exec_t *trigger;
  exec_t *action;

  if (seek == NULL)
    seek = strchr (line, '\t');
  
  if (seek){
    *seek = '\0';
  }
  else
    return;

  trigger = exec_lookup (line);
  action  = exec_lookup (seek + 1);
  *seek   = ' ';

  if (trigger == NULL || action == NULL)
    return;

  if (trigger->hook == NULL){
    trigger->hook = hook_create (action->fun);
  }
  else {
    hook_add (trigger->hook, action->fun);
  }
}



static void
add_new_list (char *line)
{
  entry_t   *entry;
  char      *tmp;
  rstring_t *item;
  
  tmp = strchr (line, ' ');

  if (tmp == NULL)
    tmp = strchr (line, '\t');

  if (tmp){
    *tmp = '\0';    /* it is replaced back with space at end of function */
    tmp++;
  }
  else
    return;

  entry = htable_insert (list_table, line, NULL);

  if (entry->content == NULL){
    item           = rstring_create ();
    entry->content = item;
  }
  item = (rstring_t *) entry->content;
  rstring_add (item, expand (tmp));

  if (tmp)
    *(tmp - 1) = ' '; /* tmp++ */
}



/*
  Empty lines, and those begining with # are omitted.  All leading
  spaces and tabs are ignored.  Then we're looking for a keyword.
*/
static void
process_line (char *buf){
  static unsigned  lineno = 0;
  char            *tmp;
  

  lineno++;

  while (isspace (*buf))
    buf++;

  if (*buf == '#' || *buf == '\0')
    return;

  tmp = strchr (buf, ' ');
  if (tmp == NULL)
    tmp = strchr (buf, '\t');
  if (!tmp){
    error_ (0, 0, _("%s:%d: parse error"), config_filename, lineno);
    return;
  }

  *tmp = '\0';
  if (!strcmp (buf, "set"))
    add_new_variable (tmp + 1);
  else if (!strcmp (buf, "key"))
    add_new_key (tmp + 1);
  else if (!strcmp (buf, "hook"))
    add_new_hook (tmp + 1);
  else if (!strcmp (buf, "list"))
    add_new_list (tmp + 1);
  else
    error_ (0, 0, "%s:%d: parse error", config_filename, lineno);
  *tmp = ' ';
}



static void
destroy_conf_item (void *ptr)
{
  struct conf_t *conf = (struct conf_t *) ptr;

  if (conf->type == TYPE_TXT)
    xfree (conf->value.txt);
  xfree (conf);
}



static void
destroy_list_item (void *ptr)
{
  rstring_t *item = (rstring_t *) ptr;

  rstring_free_strings (item);
  rstring_delete (item);
}



/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/


/*
  Init config reader.  Give expected amount of keys.
*/

void
cf_init (void){
  char *file;
  
  if (initialized){
    error_ (0, 0, _("config reader already initialized"));
    return;
  }
  conf_table  = htable_create (4);
  list_table  = htable_create (4);
  initialized = 1;

  if (config_file == NULL)
    file = file_with_dir (getenv ("HOME"), CONF_FILE);
  else
    file = xstrdup (config_file);
  if (cf_read_file (file)){
    initialized = 0;
    htable_destroy (conf_table, destroy_conf_item);
    htable_destroy (list_table, destroy_list_item);
    conf_table = NULL;
    list_table = NULL;
  }
  xfree (file);
  file = NULL;
}




/*
  Open config file, read it and put apropiate values into
  hashtable, so that they will be easy to get.

  We're going to fill buffer with text.  It will be processed till
  last newline sign.
*/

int
cf_read_file (char *filename){
  FILE     *fp;
  char     *buffer;
  int       still_to_come = 0;
  unsigned  ret;

  config_filename = xstrdup (filename);
  
  if (!initialized){
    error_ (0, 0, _("config reader not initialized"));
    xfree (config_filename);
    return 1;
  }
  
  fp = fopen (filename, "r");
  if (!fp){
    error_ (0, errno, filename);
    xfree (config_filename);
    return 1;
  }

  buffer = xcalloc (BUFFER_SIZE, 1);

  do {
    char *newline;
    char *nlbuf;
    
    ret = fill_buffer (buffer, BUFFER_SIZE, fp);

    if (ret == -1){
      error_ (0, 0, _("incorrect format of config file %s"), filename);
      xfree (config_filename);
      xfree (buffer);
      fclose (fp);
      return 1;
    }
    else if (ret == BUFFER_SIZE - 1)
      still_to_come   = 1;
    else {
      buffer[ret]     = '\n';
      buffer[ret + 1] = '\0';
      still_to_come   = 0;
    }

    nlbuf = buffer;

    while (1){
      newline = strchr (nlbuf, '\n');
      if (!newline || newline - buffer > ret)
        break;
      *newline = '\0';
      process_line (nlbuf);
      *newline = '\n';
      nlbuf    = newline + 1;
    }
  }
  while (still_to_come);

  fclose (fp);
  xfree (config_filename);
  xfree (buffer);

  config_filename = NULL;
  buffer          = NULL;
  return 0;
}



/*
  Return NULL on failure.
*/

char *
cf_txtvar_lookup (char *name){
  entry_t       *entry;
  struct conf_t *conf;

  if (!initialized){
    return NULL;
  }
  
  entry = htable_lookup (conf_table, name);
  if (entry){
    conf = (struct conf_t *) entry->content;
    if (conf->type == TYPE_TXT)
      return conf->value.txt;
  }
  return NULL;
}




/*
  Return -1 on failure.
*/

int
cf_intvar_lookup (char *name){
  entry_t       *entry;
  struct conf_t *conf;

  if (!initialized){
    return -1;
  }
  
  entry = htable_lookup (conf_table, name);
  if (entry){
    conf = (struct conf_t *) entry->content;
    if (conf->type == TYPE_NUM)
      return conf->value.num;
  }
  return -1;
}



rstring_t *
cf_listvar_lookup (char *name)
{
  entry_t   *entry;
  rstring_t *result;

  if (!initialized)
    return NULL;
  
  entry = htable_lookup (list_table, name);
  if (entry){
    result = (rstring_t *) entry->content;
    return result;
  }
  return NULL;
}



void
cf_free_resources (void){
  if (!initialized)
    return;
  
  htable_destroy (conf_table, destroy_conf_item);
  htable_destroy (list_table, destroy_list_item);
}

/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE confread.c
 *
 ****************************************************************************/
