/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2002, 2003 rzyjontko
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/

#define _GNU_SOURCE 1

/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <signal.h>

#ifdef HAVE_GETOPT_H
# include <getopt.h>
#endif

#ifdef HAVE_LOCALE_H
# include <locale.h>
#endif

#include "ecurses.h"
#include "confread.h"
#include "status.h"
#include "folder.h"
#include "cmd.h"
#include "hash.h"
#include "address.h"
#include "clock.h"
#include "mailreader.h"
#include "screen.h"
#include "mail.h"
#include "debug.h"
#include "error.h"
#include "elmo.h"
#include "pop.h"
#include "networking.h"
#include "wrapbox.h"
#include "abook.h"
#include "read.h"
#include "exec.h"
#include "gettext.h"
#include "mybox.h"
#include "bayes.h"
#include "token.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

#ifdef HAVE_GETOPT_LONG

static struct option const long_options[] =
{
  {"help", no_argument, 0, 'h'},
  {"version", no_argument, 0, 'V'},
  {NULL, required_argument, 0, 'F'},
  {NULL, 0, NULL, 0}
};

# define GETOPT(a,b,c,d,e) getopt_long (a, b, c, d, e)
#else
# define GETOPT(a,b,c,d,e) getopt (a, b, c)
#endif

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/

htable_t *address_table = NULL;

char *config_file  = NULL;

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/

static void usage (int status);
static int  decode_switches (int argc, char **argv);
static void setup_window (void);
static void destroy_content (void *);

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static int
decode_switches (int argc, char **argv)
{
  int c;
  int ala;

  while ((c = GETOPT (argc, argv, 
                      "F:"  /* alternative config file */
                      "h"	/* help */
                      "V",	/* version */
                      long_options, &ala)) != EOF){
    
    switch (c){

      case 'F':
        config_file = optarg;
        break;

      case 'V':
        printf ("%s %s\n", PACKAGE, VERSION);
        exit (EXIT_SUCCESS);

      case 'h':
        usage (EXIT_SUCCESS);

      default:
        usage (EXIT_FAILURE);
    }
  }
         
  return optind;
}         




static void
usage (int status)
{
  printf (_("%s - ELectronic Mail Operator\n"), PACKAGE);
  printf (_("Usage: %s [OPTION]... \n"), PACKAGE);
  printf (_("Options:\n"));
  printf (_("  -F <file>        use alternate rc file\n"));
  printf (_("  -h, --help       display this help and exit\n"));
  printf (_("  -V, --version    output version information and exit\n"));
  exit (status);
}




static void
destroy_content (void *ptr)
{
  address_t *addr = (address_t *)ptr;

  address_destroy (addr);
}




static void
setup_window (void)
{
  
  signal (SIGINT, elmo_finish);
  signal (SIGTERM, elmo_finish);

  address_table = htable_create (8);
  
  if (!initscr ()){
    fprintf (stderr, _("ncurses: initialization failure\n"));
    exit (EXIT_FAILURE);
  }
  keypad (stdscr, TRUE);
  leaveok (stdscr, TRUE);
  nocbreak ();
  noecho ();
  nonl ();
  curs_set (0);

  if (has_colors ()){
    start_color ();
    init_pair (1, COLOR_WHITE, COLOR_BLUE);	// statusbar, clock
    init_pair (2, COLOR_BLACK, COLOR_CYAN);	// select bar (when selected)
    init_pair (3, COLOR_RED, COLOR_CYAN);	// progress bar
    init_pair (4, COLOR_RED, COLOR_BLACK);	// message header in message view mode
    init_pair (5, COLOR_CYAN, COLOR_BLACK);	
    init_pair (6, COLOR_GREEN, COLOR_BLACK);
    init_pair (7, COLOR_YELLOW, COLOR_BLACK);
    init_pair (8, COLOR_WHITE, COLOR_BLACK);
  }

  debug_start ();
  status_start ();
  status_put_str (_("     initializing..."));
  exec_init ();
  cmd_init ();
  error_init ();
  clock_init ();
  cf_init ();
  abook_start ();
  token_init ();
  folder_init ();
  screen_doupdate ();
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/


int
main (int argc, char **argv)
{
  int i;

#if ENABLE_NLS
  setlocale (LC_ALL, "");
  bindtextdomain(PACKAGE, LOCALEDIR);
  textdomain(PACKAGE);
#endif

  i = decode_switches (argc, argv);
  srand (time (NULL));
  setup_window ();

  cmd_read_loop ();

  elmo_finish (0);
  exit (EXIT_SUCCESS);
}



void
elmo_finish (int signum)
{
  clock_free_resources ();
  mybox_free_resources ();
  net_free_resources ();
  pop_free_resources ();
  read_free_resources ();
  abook_end ();
  folder_free_resources ();
  bayes_free_resources ();
  exec_free_resources ();
  wrapbox_free_resources ();
  mail_free_resources ();
  htable_destroy (address_table, destroy_content);
  cf_free_resources ();
  error_free_resources ();
  status_free_resources ();
  screen_free_resources ();
  curs_set(1);
  debug_end ();
  if (signum)
    exit (EXIT_SUCCESS);
}


/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE elmo.c
 *
 ****************************************************************************/
