/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2002, 2003 rzyjontko
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   handle mail folder
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <string.h>

#include "ecurses.h"
#include "folder.h"
#include "xmalloc.h"
#include "line.h"
#include "screen.h"
#include "multree.h"
#include "status.h"
#include "error.h"
#include "file.h"
#include "ask.h"
#include "fetch.h"
#include "cmd.h"
#include "wrapbox.h"
#include "bayes.h"
#include "progress.h"
#include "mybox.h"
#include "hook.h"
#include "exec.h"
#include "eprintf.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define MIN(a,b) (((a)<(b))?(a):(b))

#define PREAMBLE do { if (mail_count == 0) return; } while (0)

typedef enum {ORDER_DATE, ORDER_DATE_REV, ORDER_THREAD,
              ORDER_FROM, ORDER_FROM_REV, ORDER_SUBJECT, ORDER_SUBJECT_REV}
current_order_t;

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

/**
 * There may be few pages filled with list of mails.  pages holds array
 * of WINDOW pointers.  current_page_index holds page which should be on
 * the top of visibility.  current_page_index is recalculated in set_bar.
 */
static WINDOW **pages              = NULL;
static int      current_page_index = 0;

/**
 * page_size holds number of lines in single WINDOW object
 * page_count holds size of pages array
 */
static int page_size  = 0;
static int page_count = 0;



/**
 * Bar is a hilighted position.  bar holds this window (which is then
 * moved) and bar_index holds index in mail_array which is currently chosen.
 */
static WINDOW *bar       = NULL;
static int     bar_index = 0;


/**
 * This is used by rehash_mail_array for ordering mails by threads.
 */
static multree_t  *root           = NULL;
static multree_t **nodes_array    = NULL;
static unsigned    preorder_index = 0;


/**
 * This array is used when sorting mails in threads.  Mails are copied into
 * it in proper order, and then mail_array is switched to it.
 */
static mail_t *new_mail_array = NULL;


/**
 * This is for displaing on status line.
 */
static char *boxname = NULL;


static current_order_t current_order = ORDER_THREAD;


/**
 * This is where we put the string containing single line.
 */
static char *line_txt = NULL;


/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/

static void fill_page (int i);
static void free_resources (void);
static void set_window_succession (void);
static void set_bar (void);
static void preorder_fun (multree_t *node);
static void prepare_nodes_array (void);
static int  mail_array_index (int index);
static int  first_unread_index (void);
  
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/


static void
set_color (chtype *str, int color)
{
  while (*str){
    *str |= color;
    str++;
  }
}



static void
unset_color (chtype *str, int color)
{
  while (*str){
    *str &= ~color;
    str++;
  }
}


static void
print_line (int index, WINDOW *win, int y, int x, int color)
{
  char        *txt;
  static char *fmt = NULL;

  if (fmt == NULL)
    fmt = ask_noninteractive ("line_format");

  if (fmt == NULL)
    fmt = "%?%$%# %D %x %S  ";

  if (line_txt == NULL)
    line_txt = xmalloc (COLS + 2);
  
  txt = eprintf_mail (fmt, mail_array + index, line_txt);
  
  mvwaddstr (win, y, x, txt);
  if (current_order == ORDER_THREAD){
    set_color (mail_array[index].tree, color);
    waddchstr (win, mail_array[index].tree);
    unset_color (mail_array[index].tree, color);
    getyx (win, y, x);
    wmove (win, y, x + mail_array[index].tree_len - 1);
  }
  getyx (win, y, x);
  if (mail_array[index].subject)
    waddnstr (win, mail_array[index].subject, COLS - x - 1);
  wclrtoeol (win);
}


static void
copy_tree (chtype *to, chtype *from)
{
  while (*from && *from != '>'){
    if (*from == ' ' || *from == ACS_VLINE)
      *to = *from;
    else if (*from == ACS_LLCORNER || *from == ACS_HLINE)
      *to = ' ';
    else if (*from == ACS_LTEE)
      *to = ACS_VLINE;
    else
      *to = *from;
    to++;
    from++;
  }
}



static void
preorder_fun (multree_t *node)
{
  mail_t    *self;
  mail_t    *parent;
  multree_t *parent_node;
  
  if (node == root)
    return;

  self = &mail_array[(unsigned)node->data];
  if (self->tree)
    xfree (self->tree);
  if (self->parent != -1){
    parent_node    = nodes_array[self->parent];
    parent         = &mail_array[self->parent];
    self->tree_len = parent->tree_len + ((parent->tree_len) ? 2 : 4);
    self->tree     = xmalloc (self->tree_len * sizeof (chtype));
    self->tree[self->tree_len - 3] = ACS_HLINE;
    self->tree[self->tree_len - 2] = '>';
    self->tree[self->tree_len - 1] = 0;
    copy_tree (self->tree, parent->tree);
    if (parent_node->children[parent_node->children_filled - 1] == node)
      self->tree[self->tree_len - 4] = ACS_LLCORNER;
    else
      self->tree[self->tree_len - 4] = ACS_LTEE;
  }
  else {
    parent_node    = NULL;
    parent         = NULL;
    self->tree_len = 0;
    self->tree     = xcalloc (1, sizeof (chtype));
  }

  mail_copy_to_array (new_mail_array, (unsigned)node->data, preorder_index);
  preorder_index++;
}



static int
date_cmp (const void *a, const void *b)
{
  mail_t *aa = (mail_t *) a;
  mail_t *bb = (mail_t *) b;

  return (aa->date < bb->date) - (aa->date > bb->date);
}



static int
date_cmp_rev (const void *a, const void *b)
{
  mail_t *aa = (mail_t *) a;
  mail_t *bb = (mail_t *) b;

  return (aa->date > bb->date) - (aa->date < bb->date);
}



static int
from_cmp (const void *a, const void *b)
{
  mail_t *aa = (mail_t *) a;
  mail_t *bb = (mail_t *) b;

  return strcoll (aa->from->name, bb->from->name);
}



static int
from_cmp_rev (const void *a, const void *b)
{
  mail_t *aa = (mail_t *) a;
  mail_t *bb = (mail_t *) b;

  return strcoll (bb->from->name, aa->from->name);
}



static int
subject_cmp (const void *a, const void *b)
{
  mail_t *aa = (mail_t *) a;
  mail_t *bb = (mail_t *) b;
  
  if ((!aa) && (!bb)) 
    return strcoll (aa->subject, bb->subject);
  else
     return 0;
}



static int
subject_cmp_rev (const void *a, const void *b)
{
  mail_t *aa = (mail_t *) a;
  mail_t *bb = (mail_t *) b;

  if ((!aa) && (!bb)) 
    return strcoll (bb->subject, aa->subject);
  else
     return 0;
}



static int
index_cmp (const void *a, const void *b)
{
  mail_t *aa = (mail_t *) a;
  mail_t *bb = (mail_t *) b;

  return (aa->index > bb->index) - (aa->index < bb->index);
}



static int
mail_array_index (int index)
{
  int i;

  for (i = 0; i < mail_count; i++)
    if (mail_array[i].index == index)
      break;
  return i;
}



static void
make_date_order (void)
{
  int old_index = mail_array[bar_index].index;
  
  if (current_order != ORDER_DATE){
    qsort (mail_array, mail_count, sizeof (mail_t), date_cmp);
    current_order = ORDER_DATE;
  }
  else {
    qsort (mail_array, mail_count, sizeof (mail_t), date_cmp_rev);
    current_order = ORDER_DATE_REV;
  }
  bar_index = mail_array_index (old_index);
}



static void
make_from_order (void)
{
  int old_index = mail_array[bar_index].index;
  
  if (current_order != ORDER_FROM){
    qsort (mail_array, mail_count, sizeof (mail_t), from_cmp);
    current_order = ORDER_FROM;
  }
  else {
    qsort (mail_array, mail_count, sizeof (mail_t), from_cmp_rev);
    current_order = ORDER_FROM_REV;
  }
  bar_index = mail_array_index (old_index);
}



static void
make_subject_order (void)
{
  int old_index = mail_array[bar_index].index;
  
  if (current_order != ORDER_SUBJECT){
    qsort (mail_array, mail_count, sizeof (mail_t), subject_cmp);
    current_order = ORDER_SUBJECT;
  }
  else {
    qsort (mail_array, mail_count, sizeof (mail_t), subject_cmp_rev);
    current_order = ORDER_SUBJECT_REV;
  }
  bar_index = mail_array_index (old_index);
}



static void
make_index_order (void)
{
  int old_index = mail_array[bar_index].index;
  
  qsort (mail_array, mail_count, sizeof (mail_t), index_cmp);
  bar_index = mail_array_index (old_index);
}



/**
 * This is how it works.
 * First I have to sort mail_array according to index.
 * There is a special array nodes_array which holds pointers to multrees --
 * -- one pointer for each mail.  Every node of multree holds its index in
 * mail_array and array of pointers to its children (replies).
 * Then I just go in preorder manner and put mails into new_mail_array, and
 * finally I switch to that table.
 */
static void
make_thread_order (void)
{
  int        i;
  multree_t *parent;
  int old_index = mail_array[bar_index].index;
  
  make_index_order ();
  root           = multree_create (mail_not_replies_count);
  root->data     = (void *) -1;
  nodes_array    = xcalloc (mail_count, sizeof (multree_t *));
  new_mail_array = mail_new_array ();
  prepare_nodes_array ();
  for (i = 0; i < mail_count; i++){
    if (mail_array[i].parent == -1)
      parent = root;
    else
      parent = nodes_array[mail_array[i].parent];
    multree_add_child (parent, nodes_array[i]);
  }
  preorder_index = 0;
  multree_preorder (root, preorder_fun);
  multree_postorder (root, multree_node_destroy);
  mail_switch_to_array (new_mail_array);
  if (nodes_array)
    xfree (nodes_array);
  root           = NULL;
  nodes_array    = NULL;
  new_mail_array = NULL;
  bar_index      = mail_array_index (old_index);
  current_order  = ORDER_THREAD;
}



static void
prepare_nodes_array (void)
{
  int i;

  for (i = 0; i < mail_count; i++){
    nodes_array[i]       = multree_create (mail_array[i].child_count);
    nodes_array[i]->data = (void *) i;
  }
}



static void
fill_page (int num)
{
  WINDOW *page  = pages[num];
  int     index = num * page_size;
  int     i;

  wclear (page);
  for (i = 0; i < page_size; index++, i++){
    if (index >= mail_count)
      break;
    print_line (index, page, i, 0, COLOR_PAIR (4));
  }
}


static void
free_resources (void)
{
  int i;

  if (line_txt){
    xfree (line_txt);
    line_txt = NULL;
  }
  
  if (pages == NULL)
    return;

  for (i = 0; i < page_count; i++)
    delwin (pages[i]);
  xfree (pages);
  pages = NULL;

  if (bar)
    delwin (bar);
  bar = NULL;
}



static void
set_window_succession (void)
{
  set_bar ();
  
  if (pages){
    screen_touchwin (pages[current_page_index]);
    screen_wnoutrefresh (pages[current_page_index]);
    screen_touchwin (bar);
    screen_wnoutrefresh (bar);
  }
}




static void
set_bar (void)
{
  if (bar_index < 0 || bar_index >= mail_count){
    current_page_index = 0;
    return;
  }

  current_page_index = bar_index / page_size;
  
  if (!bar){
    bar = newwin (1, COLS, bar_index % page_size, 0);
    wbkgdset (bar, COLOR_PAIR (2));
  }
  else {
    mvwin (bar, bar_index % page_size, 0);
  }

  print_line (bar_index, bar, 0, 0, COLOR_PAIR (3));
}




static void
put_mails_on_pages (void)
{
  int i;

  if (pages)
    free_resources ();
  
  page_size  = LINES - 2;
  page_count = mail_count / page_size + 1;

  pages = xmalloc (page_count * sizeof (WINDOW *));
  for (i = 0; i < page_count; i++){
    pages[i] = newwin (page_size, COLS, 0, 0);
    fill_page (i);
    screen_touchwin (pages[i]);
    screen_wnoutrefresh (pages[i]);
  }
}




static int
first_unread_index (void)
{
  int i;

  for (i = 0; i < mail_count; i++)
    if (!(mail_array[i].flags & FLAG_READ))
      return i;
  return i - 1;
}




static void
display_status_info (char *name)
{
  if (name != boxname){
    if (boxname)
      xfree (boxname);
    boxname = xstrdup (name);
  }
  status_put_mailbox (boxname, mail_count, mail_old_count - mail_read_count,
                      mail_count - mail_old_count);
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

int
folder_bar_index (void)
{
  return bar_index;
}




void
folder_init (void)
{
  char *box;

  box = mybox_start ();
  
  folder_read_box (box);
  xfree (box);
}




void
folder_free_resources (void)
{
  folder_at_leave ();
  
  wrapbox_free_resources ();
  free_resources ();
  if (boxname){
    xfree (boxname);
    boxname = NULL;
  }
}




void
folder_sort_threads (void)
{
  PREAMBLE;
  
  make_thread_order ();
  put_mails_on_pages ();
  set_window_succession ();
}



void
folder_sort_date (void)
{
  PREAMBLE;
  
  make_date_order ();
  put_mails_on_pages ();
  set_window_succession ();
}



void
folder_sort_from (void)
{
  PREAMBLE;
  
  make_from_order ();
  put_mails_on_pages ();
  set_window_succession ();
}



void
folder_sort_subject (void)
{
  PREAMBLE;
  
  make_subject_order ();
  put_mails_on_pages ();
  set_window_succession ();
}



void
folder_sort_index (void)
{
  PREAMBLE;
  
  make_index_order ();
  put_mails_on_pages ();
  set_window_succession ();
}



int
folder_read_box (char *name)
{
  mail_reset_array ();
  
  if (wrapbox_read_box (name) < 0){
    error_ (0, 0, "%s is not a valid mailbox", name);
    return 1;
  }

  display_status_info (name);

  free_resources ();

  bar_index = first_unread_index ();
  folder_sort_threads ();
  return 0;
}



void
folder_page_next (void)
{
  if (current_page_index == page_count - 1)
    bar_index = mail_count - 1;
  else
    bar_index = MIN (bar_index + page_size, mail_count - 1);
  set_window_succession ();
}



void
folder_page_prev (void)
{
  if (current_page_index == 0)
    bar_index = 0;
  else
    bar_index -= page_size;
  set_window_succession ();
}



void
folder_bar_next (void)
{
  if (bar_index == mail_count - 1)
    return;
  bar_index++;
  set_window_succession ();
}




void
folder_bar_prev (void)
{
  if (bar_index == 0)
    return;
  bar_index--;
  set_window_succession ();
}



void
folder_bar_first (void)
{
  bar_index = 0;
  set_window_succession ();
}



void
folder_bar_last (void)
{
  bar_index = mail_count - 1;
  set_window_succession ();
}



void
folder_soft_refresh (void)
{
  set_window_succession ();
  status_put_mailbox (boxname, mail_count, mail_old_count - mail_read_count,
                      mail_count - mail_old_count);
}


void
folder_refresh (void)
{
  int i;
  
  if (pages){
    for (i = 0; i < page_count; i++){
      fill_page (i);
      screen_wnoutrefresh (pages[i]);
    }
    set_window_succession ();
    screen_redrawwin (pages[current_page_index]);
    if (bar)
      screen_redrawwin (bar);
    status_put_mailbox (boxname, mail_count, mail_old_count - mail_read_count,
                        mail_count - mail_old_count);
  }
}



void
folder_re_read (void)
{
  folder_read_box (boxname);
  folder_refresh ();
}



void
folder_redraw_bar (void)
{
  if (cmd_state_get () == CMD_LIST){
    print_line (bar_index, pages[current_page_index],
                bar_index % page_size, 0, COLOR_PAIR (4));
    set_window_succession ();
  }
}



void
folder_after_read (void)
{
  PREAMBLE;
  
  if (!(mail_array[bar_index].flags & FLAG_READ)){
    mail_read_count++;
    if (!(mail_array[bar_index].flags & FLAG_OLD))
      mail_old_count++;
  }
  
  wrapbox_add_flag (mail_array + bar_index, FLAG_OLD);
  wrapbox_add_flag (mail_array + bar_index, FLAG_READ);
  folder_redraw_bar ();
}



void
folder_after_reply (void)
{
  PREAMBLE;
  
  folder_after_read ();
  wrapbox_add_flag (mail_array + bar_index, FLAG_ANSWERED);
  folder_redraw_bar ();
}



void
folder_after_fwd (void)
{
  PREAMBLE;
  
  folder_after_read ();
  wrapbox_add_flag (mail_array + bar_index, FLAG_PASSED);
  folder_redraw_bar ();
}



void
folder_delete_mail (void)
{
  PREAMBLE;
  
  wrapbox_add_flag (mail_array + bar_index, FLAG_TRASHED);
  folder_redraw_bar ();
}



void
folder_toggle_flag (void)
{
  PREAMBLE;
  
  if (mail_array[bar_index].flags & FLAG_FLAGGED)
    wrapbox_remove_flag (mail_array + bar_index, FLAG_FLAGGED);
  else
    wrapbox_add_flag (mail_array + bar_index, FLAG_FLAGGED);
  folder_redraw_bar ();
}



void
folder_flag_all (void)
{
  int i;

  for (i = 0; i < mail_count; i++){
    wrapbox_add_flag (mail_array + i, FLAG_FLAGGED);
  }
  folder_refresh ();
}



void
folder_unflag_all (void)
{
  int i;

  for (i = 0; i < mail_count; i++){
    wrapbox_remove_flag (mail_array + i, FLAG_FLAGGED);
  }
  folder_refresh ();
}




void
folder_flag_invert (void)
{
  int i;

  for (i = 0; i < mail_count; i++){
    if (mail_array[i].flags & FLAG_FLAGGED)
      wrapbox_remove_flag (mail_array + i, FLAG_FLAGGED);
    else
      wrapbox_add_flag (mail_array + i, FLAG_FLAGGED);
  }
  folder_refresh ();
}



void
folder_flush (void)
{
  wrapbox_remove_trashed ();
  folder_re_read ();
}



void
folder_move_mail (void)
{
  char *box;
  char *autoflush = ask_noninteractive ("autoflush");

  PREAMBLE;
  
  box = mybox_select_subdir ();

  if (box == NULL)
    return;

  wrapbox_move_mail_to (mail_array + bar_index, box);
  xfree (box);
  if(autoflush)
    folder_flush();
}



void
folder_move_flagged (void)
{
  int   i;
  char *box = mybox_select_subdir ();

  if (box == NULL)
    return;

  for (i = 0; i < mail_count; i++){
    if (mail_array[i].flags & FLAG_FLAGGED){
      wrapbox_move_mail_to (mail_array + i, box);
    }
  }
  xfree (box);
}



void
folder_flagged_make_read (void)
{
  int i;

  for (i = 0; i < mail_count; i++){
    if (mail_array[i].flags & FLAG_FLAGGED){
      wrapbox_add_flag (mail_array + i, FLAG_OLD);
      wrapbox_add_flag (mail_array + i, FLAG_READ);
    }
  }
}



void
folder_check_new_for_spam (void)
{
  int i;

  for (i = 0; i < mail_count; i++){
    if ((mail_array[i].flags & FLAG_READ)
        || (mail_array[i].flags & FLAG_OLD))
      continue;
    if (bayes_is_spam (mail_array + i))
      wrapbox_add_flag (mail_array + i, FLAG_SPAM);
    else
      wrapbox_add_flag (mail_array + i, FLAG_NOSPAM);
  }
  folder_refresh ();
}



void
folder_check_not_checked (void)
{
  int i;

  progress_open ("scanning");
  progress_set_max (mail_count);
  for (i = 0; i < mail_count; i++){
    if (! (mail_array[i].flags & FLAG_SPAM)
        && ! (mail_array[i].flags & FLAG_NOSPAM)
        && ! (mail_array[i].flags & FLAG_LEGITIMATE)){
      if (bayes_is_spam (mail_array + i))
        wrapbox_add_flag (mail_array + i, FLAG_SPAM);
      else
        wrapbox_add_flag (mail_array + i, FLAG_NOSPAM);
    }
    progress_add (1);
  }
  progress_close ();
  folder_refresh ();
}



void
folder_check_spam (void)
{
  PREAMBLE;
  
  if (bayes_is_spam (mail_array + bar_index))
    wrapbox_add_flag (mail_array + bar_index, FLAG_SPAM);
  else
    wrapbox_add_flag (mail_array + bar_index, FLAG_NOSPAM);
  folder_redraw_bar ();
}



void
folder_delete_as_spam (void)
{
  PREAMBLE;
  
  if (mail_array[bar_index].flags & FLAG_LEGITIMATE){
    bayes_unscan_legitimate (mail_array + bar_index);
  }
  wrapbox_remove_flag (mail_array + bar_index, FLAG_NOSPAM);
  wrapbox_add_flag (mail_array + bar_index, FLAG_SPAM);
  wrapbox_add_flag (mail_array + bar_index, FLAG_TRASHED);
  folder_redraw_bar ();
}



void
folder_make_legitimate (void)
{
  PREAMBLE;
  
  bayes_scan_legitimate (mail_array + bar_index);
  wrapbox_add_flag (mail_array + bar_index, FLAG_LEGITIMATE);
  folder_redraw_bar ();
}



void
folder_trash_spam (void)
{
  int i;

  for (i = 0; i < mail_count; i++){
    if (mail_array[i].flags & FLAG_SPAM){
      bayes_scan_spam (mail_array + i);
      wrapbox_add_flag (mail_array + i, FLAG_TRASHED);
    }
  }
}


void
folder_scan_legitimate (void)
{
  int i;

  for (i = 0; i < mail_count; i++){
    if (! (mail_array[i].flags & FLAG_LEGITIMATE)
        && (mail_array[i].flags & FLAG_NOSPAM)){
      bayes_scan_legitimate (mail_array + i);
      wrapbox_add_flag (mail_array + i, FLAG_LEGITIMATE);
    }
  }
}


void
folder_none_was_scanned (void)
{
  int i;

  for (i = 0; i < mail_count; i++){
    wrapbox_remove_flag (mail_array + i, FLAG_LEGITIMATE);
  }
}


void
folder_at_leave (void)
{
  exec_t *exec = exec_lookup_fun (folder_at_leave);

  hook_execute (exec->hook);
}

/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE folder.c
 *
 ****************************************************************************/
